#include "WeArtController.h"
#include "weArtPlugin.h"
#include "Async/Async.h"

#include "Networking.h"
#include "Sockets.h"
#include "SocketSubsystem.h"

#include "WeArtSettings.h"
#include "CoreMinimal.h"
#include "Engine/GameEngine.h"

UWeArtController::UWeArtController() : trailingText()
{
}

UWeArtController::~UWeArtController()
{
}

void UWeArtController::Initialize(FSubsystemCollectionBase& Collection)
{
	//Get the settings from Edit/Project Settings/Plugins/WeArt
	UWeArtSettings* Plugin_Settings = GetMutableDefault<UWeArtSettings>();

	ClientPort = Plugin_Settings->ClientPort;
	DebugMessages = Plugin_Settings->DebugMessages;
	ControllerAutoStart = Plugin_Settings->ControllerAutoStart;
	CalibrationAutoStart = Plugin_Settings->CalibrationAutoStart;

	IsPaused = ! ControllerAutoStart;

	if (ControllerAutoStart) {
		ScheduleController();
	}

	TArray<AActor*> FoundActors;
	UGameplayStatics::GetAllActorsOfClass(GetWorld(), AWeArtTrackingCalibration::StaticClass(), FoundActors);
	
	if (FoundActors.Num() > 0)
	{
		foundTrackingCalibration = true;
		TrackingCalibration = Cast<AWeArtTrackingCalibration>(FoundActors[0]);
		TrackingCalibration->OnMiddlewareCalibrationStart.AddUniqueDynamic(TrackingCalibration, &AWeArtTrackingCalibration::OnCalibrationStart);
		TrackingCalibration->OnMiddlewareCalibrationFinish.AddUniqueDynamic(TrackingCalibration, &AWeArtTrackingCalibration::OnCalibrationFinish);
		TrackingCalibration->OnMiddlewareCalibrationResultSuccess.AddUniqueDynamic(TrackingCalibration, &AWeArtTrackingCalibration::OnCalibrationResultSuccess);
		TrackingCalibration->OnMiddlewareCalibrationResultFail.AddUniqueDynamic(TrackingCalibration, &AWeArtTrackingCalibration::OnCalibrationResultFail);
		TrackingCalibration->OnMiddlewareCalibrationStop.AddUniqueDynamic(TrackingCalibration, &AWeArtTrackingCalibration::OnCalibrationStop);
	}
}

void UWeArtController::Deinitialize()
{
	// This send does not work, why?
	StopFromClientMessage stopMsg;
	SendMessage(&stopMsg);

	StopConnection();
}


void UWeArtController::PauseController()
{
	if (IsPaused) {
		return;
	}

	IsPaused = true;
	Deinitialize();
}


void UWeArtController::UnpauseController()
{
	if (!IsPaused) {
		return;
	}

	IsPaused = false;
	ScheduleController();
}

void UWeArtController::StartCalibration()
{
	StartCalibrationMessage startCalibrationMessage;
	SendMessage(&startCalibrationMessage);
}

void UWeArtController::StopCalibration()
{
	StopCalibrationMessage stopCalibrationMessage;
	SendMessage(&stopCalibrationMessage);
	TrackingCalibration->OnMiddlewareCalibrationStop.Broadcast();
}

void UWeArtController::SendMessage(WeArtMessage* message)
{
	if (!IsConnected) {
		return;
	}

	// WeArt message to string
	std::string text = messageSerializer.Serialize(message);
	text += messagesSeparator;

	// String to bytestream
	size_t textSize = text.size();
	FString fText(text.c_str());
	TCHAR* bytestream = fText.GetCharArray().GetData();
	int32 size = FCString::Strlen(bytestream);

	// Send the message
	int32 sentAmount = 0;
	if (socket->Send((uint8*)TCHAR_TO_ANSI(bytestream), size, sentAmount)) {
		FString fMessage(text.c_str());
		LogMessage(WeArtMessageType::MessageSent, fMessage);
	}
	else {
		FString fMessage(text.c_str());
		LogError(WeArtError::SendMessageError, fMessage);
	}
}

bool UWeArtController::ReceiveMessages(std::vector<WeArtMessage*>& messages)
{
	if (!IsConnected) {
		return false;
	}

	TArray<uint8> receivedData;

	uint32 size;
	while (socket->HasPendingData(size))
	{
		receivedData.SetNumUninitialized(FMath::Min(size, 65507u));
		int32 Read = 0;
		socket->Recv(receivedData.GetData(), receivedData.Num(), Read);
	}

	if (receivedData.Num() > 0) {
		// Bytestream to one large string
		FString fBufferText = ANSI_TO_TCHAR(reinterpret_cast<const char*>(receivedData.GetData()));

		// For UE-related reasons, the above conversion can leave
		// some garbage entries at the end of the string. Cut them off here.
		if (fBufferText.Len() - receivedData.Num() > 0) {
			fBufferText.RemoveAt(receivedData.Num(), fBufferText.Len() - receivedData.Num());
		}

		std::string bufferText = TCHAR_TO_UTF8(*fBufferText);
		bufferText = trailingText + bufferText;

		const int lastSeparatorIndex = bufferText.find_last_of(messagesSeparator);
		trailingText = bufferText.substr(lastSeparatorIndex + 1);
		std::string textToParse = bufferText.substr(0, lastSeparatorIndex);

		// Split the string on separator occurence
		std::vector<std::string> splitStrings;
		std::istringstream dataStream(textToParse);
		std::string s;
		while (getline(dataStream, s, messagesSeparator)) {
			if (s.empty())
				continue;
			splitStrings.push_back(s);
		}

		// Std::strings to WeArtMessages
		messages.resize(splitStrings.size());
		for (int i = 0; i < messages.size(); i++) {
			messages[i] = messageSerializer.Deserialize(splitStrings[i]);
			FString fMessage(splitStrings[i].c_str());
			LogMessage(WeArtMessageType::MessageReceived, fMessage);
		}
		return true;
	}
	else {
		messages = std::vector<WeArtMessage*>();
		return false;
	}
}

void UWeArtController::OnMessageReceived(WeArtMessage* msg)
{
	// Currently, only exit/disconnect messages
	// are supported. More might come in the future.
	std::string ID = msg->getID();
	if (ID == "exit" || ID == "disconnect") {
		StopConnection();
	}

}


void UWeArtController::StopConnection()
{
	IsConnected = false;
	isCancellationRequested = true;
	if (socket) {
		socket->Close();
	}
}


void UWeArtController::LogMessage(WeArtMessageType type, FString& message)
{
	if (type == WeArtMessageType::MessageReceived)
	{
		HandleReceivedMessage(message);
	}

	if (!DebugMessages) {
		return;
	}

	switch (type) {
	case WeArtMessageType::MessageSent: {
		UE_LOG(LogWeArt, Log, TEXT("To Middleware: %s"), *message);
		break;
	}
	case WeArtMessageType::MessageReceived: {
		UE_LOG(LogWeArt, Log, TEXT("From Middleware: %s"), *message);
		break;
	}
	default:
		break;
	}
}

void UWeArtController::HandleReceivedMessage(FString& message)
{
	if (message.StartsWith("CalibrationStatus"))
	{
		TArray<FString> Out;
		message.ParseIntoArray(Out, TEXT(":"), true);

		if (Out[2] == "1")
		{
			if(foundTrackingCalibration)
			TrackingCalibration->OnMiddlewareCalibrationStart.Broadcast(Out[1] == "1" ? HandSide::Right : HandSide::Left);
		}

		if (Out[2] == "2")
		{
			if(foundTrackingCalibration)
			TrackingCalibration->OnMiddlewareCalibrationFinish.Broadcast(Out[1] == "1" ? HandSide::Right : HandSide::Left);
		}
	}

	if (message.StartsWith("CalibrationResult"))
	{
		TArray<FString> Out;
		message.ParseIntoArray(Out, TEXT(":"), true);

		if (Out[2] == "0")
		{
			if(foundTrackingCalibration)
			TrackingCalibration->OnMiddlewareCalibrationResultSuccess.Broadcast(Out[1] == "1" ? HandSide::Right : HandSide::Left);
		}
		else
		{
			if(foundTrackingCalibration)
			TrackingCalibration->OnMiddlewareCalibrationResultFail.Broadcast(Out[1] == "1" ? HandSide::Right : HandSide::Left);
		}
	}
}


void UWeArtController::LogError(WeArtError error, FString& message)
{
	if (!DebugMessages) {
		return;
	}

	switch (error) {
	case WeArtError::ConnectionError: {
		UE_LOG(LogWeArt, Error, TEXT("Cannot connect to %s."), *message);
		break;
	}
	case WeArtError::ReceiveMessageError: {
		UE_LOG(LogWeArt, Error, TEXT("Error on receiving message %s"), *message);
		break;
	}
	case WeArtError::SendMessageError: {
		UE_LOG(LogWeArt, Error, TEXT("Error on sending message %s"), *message);
		break;
	}
	default:
		break;
	};
}

void UWeArtController::ScheduleController()
{
	AsyncTask(ENamedThreads::AnyHiPriThreadNormalTask, [&]() {

		FString addrStr;

		while (!isCancellationRequested) {
			// Create a socket and address structure in a platform agnostic way.
			socket = ISocketSubsystem::Get(PLATFORM_SOCKETSUBSYSTEM)->CreateSocket(NAME_Stream, TEXT("default"), false);
			TSharedRef<FInternetAddr> addr = ISocketSubsystem::Get(PLATFORM_SOCKETSUBSYSTEM)->CreateInternetAddr();

			// Find the IP address of the local ethernet adapter.
			// The socket subsystem itself performs some checks
			// which should filter out most unwanted results.
			// If that's not enough, overriding the following
			// getter might be necessary.
			TArray<TSharedPtr<FInternetAddr>> addresses;
			ISocketSubsystem::Get(PLATFORM_SOCKETSUBSYSTEM)->GetLocalAdapterAddresses(addresses);
			if (addresses.Num() == 0) {
				assert(false);
			}
			else {
				uint32 ip;
				addresses[0]->GetIp(ip);
				addr->SetIp(ip);
				addr->SetPort(ClientPort);
				addrStr = *addr->ToString(true);
			}

			// Attempt to connect to the target.
			if (socket->Connect(*addr)) {
				UE_LOG(LogWeArt, Log, TEXT("Connected to %s."), *addrStr);
				IsConnected = true;
				StartFromClientMessage startMsg;
				/*
				std::string TestString = startMsg.getID() + ":" + startMsg.getValues()[0] + ":" + startMsg.getValues()[1] + ":" + startMsg.getValues()[2];
				FString HappyString(TestString.c_str());
				GEngine->AddOnScreenDebugMessage(-1, 15.0f, FColor::Yellow, FString::Printf(TEXT("%s "),HappyString));
				*/
				SendMessage(&startMsg);

				if (CalibrationAutoStart)
				{
					StartCalibrationMessage startCalibrationMessage;
					SendMessage(&startCalibrationMessage);
				}
			}
			else {
				FString ipString = addr->ToString(true);
				LogError(WeArtError::ConnectionError, ipString);
				StopConnection();
			}

			// Keep receiving messages unless cancelled.
			while (!isCancellationRequested) {
				std::vector<WeArtMessage*> msgs;
				bool received = ReceiveMessages(msgs);
				if (received) {
					for (auto msg : msgs) {

						if (msg == NULL)
							continue;

						OnMessageReceived(msg);

						// Forward the message to relevant tracking objects
						for (UWeArtThimbleTrackingObject* obj : thimbleTrackingObjects) {
							obj->OnMessageReceived(msg);
						}
					}
				}
			}
		}


	// Log address
	if (socket->GetConnectionState() != SCS_Connected) {
		UE_LOG(LogWeArt, Log, TEXT("Disconnected from %s."), *addrStr);
	}
	});
}
