// Fill out your copyright notice in the Description page of Project Settings.


#include "WeArtHandController.h"
#include "WeArtController.h"
#include "Components/CapsuleComponent.h"

// Sets default values for this component's properties
UWeArtHandController::UWeArtHandController()
{
	// Set this component to be initialized when the game starts, and to be ticked every frame.  You can turn these features
	// off to improve performance if you don't need them.
	PrimaryComponentTick.bCanEverTick = true;
	// ...
}

void UWeArtHandController::SetIsHandGrabInvisible(bool value)
{
	IsHandGrabInvisible = value;
}

bool UWeArtHandController::GetIsHandGrabInvisible()
{
	return IsHandGrabInvisible;
}


// Called when the game starts
void UWeArtHandController::BeginPlay()
{
	Super::BeginPlay();
	TArray<AActor*> FoundActors;
	UGameplayStatics::GetAllActorsOfClass(GetWorld(), AWeArtGraspEvents::StaticClass(), FoundActors);

	if (FoundActors.Num() > 0)
	{
		foundGraspEvents = true;
		weArtGraspEvents = Cast<AWeArtGraspEvents>(FoundActors[0]);
	}

}


void UWeArtHandController::Init(UWeArtThimbleTrackingObject* thumbTracking, UWeArtThimbleTrackingObject* indexTracking,
	UWeArtThimbleTrackingObject* middleTracking, UWeArtHapticObject* thumbHaptic, UWeArtHapticObject* indexHaptic, UWeArtHapticObject* middleHaptic)
{
	// Init haptic components
	thumbHapticObject = thumbHaptic;
	thumbHapticObject->GetOwner()->OnActorBeginOverlap.AddDynamic(this, &UWeArtHandController::BeginOverlap);
	thumbHapticObject->GetOwner()->OnActorEndOverlap.AddDynamic(this, &UWeArtHandController::EndOverlap);
	indexHapticObject = indexHaptic;
	indexHapticObject->GetOwner()->OnActorBeginOverlap.AddDynamic(this, &UWeArtHandController::BeginOverlap);
	indexHapticObject->GetOwner()->OnActorEndOverlap.AddDynamic(this, &UWeArtHandController::EndOverlap);
	middleHapticObject = middleHaptic;
	middleHapticObject->GetOwner()->OnActorBeginOverlap.AddDynamic(this, &UWeArtHandController::BeginOverlap);
	middleHapticObject->GetOwner()->OnActorEndOverlap.AddDynamic(this, &UWeArtHandController::EndOverlap);

	// Init tracking components
	thumbTrackingObject = thumbTracking;
	indexTrackingObject = indexTracking;
	middleTrackingObject = middleTracking;

	initialized = thumbHapticObject && indexHapticObject && middleHapticObject && thumbTrackingObject && indexTrackingObject && middleTrackingObject;
	if (!initialized)
		UE_LOG(LogWeArt, Error, TEXT("Hand initialization error!"));
}


void UWeArtHandController::BeginOverlap(AActor* MyOverlappedActor, AActor* OtherActor)
{
	UWeArtTouchableObject* touchableObject = OtherActor->FindComponentByClass<UWeArtTouchableObject>();

	if (touchableObject == nullptr)
		return;
	if (!touchableObject->IsGraspable)
		return;
	if (graspingState == GraspingState::Grabbed)
		return;

	if (MyOverlappedActor == thumbHapticObject->GetOwner())
		touchedObjectByThumb = touchableObject;

	 else if (MyOverlappedActor == indexHapticObject->GetOwner())
		touchedObjectByIndex = touchableObject;

	else if (MyOverlappedActor == middleHapticObject->GetOwner())
		touchedObjectByMiddle = touchableObject;

	if (touchedObjectByThumb == touchedObjectByIndex && touchedObjectByIndex == touchedObjectByMiddle)
		touchedObject = touchableObject;
	else
		touchedObject = nullptr;
}


void UWeArtHandController::EndOverlap(AActor* MyOverlappedActor, AActor* OtherActor)
{
	if (graspingState == GraspingState::Grabbed)
		return;

	if (MyOverlappedActor == thumbHapticObject->GetOwner())
		touchedObjectByThumb = nullptr;

	else if (MyOverlappedActor == indexHapticObject->GetOwner())
		touchedObjectByIndex = nullptr;

	else if (MyOverlappedActor == middleHapticObject->GetOwner())
		touchedObjectByMiddle = nullptr;

	if (touchedObjectByThumb != touchedObjectByIndex || touchedObjectByIndex != touchedObjectByMiddle)
		touchedObject = nullptr;
}


void UWeArtHandController::CheckHandClosingState(void)
{
	if (thumbTrackingObject->GetClosure() >= WeArtConstants::thresholdThumbClosure &&
		middleTrackingObject->GetClosure() >= WeArtConstants::thresholdMiddleClosure ||
		indexTrackingObject->GetClosure() >= WeArtConstants::thresholdIndexClosure)
	{
		switch (closingState) {
		case HandClosingState::Open:
			closingState = HandClosingState::Closing;
			break;
		case HandClosingState::Closing:
			closingState = HandClosingState::Closed;
			break;
		}
	}
	else
	{
		closingState = HandClosingState::Open;
	}
}


void UWeArtHandController::HandleGraspingState(void)
{
	if (closingState == HandClosingState::Closing && touchedObject && graspingState == GraspingState::Released)
	{
		graspingState = GraspingState::Grabbed;

		if (foundGraspEvents)
		{
			weArtGraspEvents->OnMiddlewareHandControllerGrasp.Broadcast(handSide, touchedObject);
		}

		thumbHapticObject->ClearEffects();
		indexHapticObject->ClearEffects();
		middleHapticObject->ClearEffects();

		touchedObject->Grab(GetOwner(), thumbHapticObject, indexHapticObject, middleHapticObject);

		// Find actor root component (we know it's a capsule collision) and hide it.
		if (IsHandGrabInvisible == true)
		{
			UCapsuleComponent* rootComponent = GetOwner()->FindComponentByClass<UCapsuleComponent>();
			rootComponent->ToggleVisibility(true);
		}
	}
	else if (closingState == HandClosingState::Open && touchedObject && graspingState == GraspingState::Grabbed)
	{
		graspingState = GraspingState::Released;

		if (foundGraspEvents)
		{
			weArtGraspEvents->OnMiddlewareHandControllerRelease.Broadcast(handSide, touchedObject);
		}

		touchedObject->Release();

		// Find actor root component (we know it's a capsule collision) and hide it.
		if (IsHandGrabInvisible == true)
		{
			UCapsuleComponent* rootComponent = GetOwner()->FindComponentByClass<UCapsuleComponent>();
			rootComponent->ToggleVisibility(true);
		}
	}
}


// Called every frame
void UWeArtHandController::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	if (!initialized)
		return;

	CheckHandClosingState();

	HandleGraspingState();

	if (touchedObject && graspingState == GraspingState::Grabbed) {
		touchedObject->ComputeDynamicGraspForce(
			thumbTrackingObject->GetClosure(),
			middleTrackingObject->GetClosure(),
			indexTrackingObject->GetClosure()
		);
	}
}

