#include "WeArtTouchableObject.h"


const WeArtImpactInfo WeArtImpactInfo::none = WeArtImpactInfo();

static UWeArtHapticObject* TryGetHapticObjectFromActor(AActor* actor)
{
	return actor->FindComponentByClass<UWeArtHapticObject>();
}

// Sets default values for this component's properties
UWeArtTouchableObject::UWeArtTouchableObject()
{
	PrimaryComponentTick.bCanEverTick = true;
}


// Called when the game starts
void UWeArtTouchableObject::BeginPlay()
{
	Super::BeginPlay();

	// Register our custom overlap events.
	AActor* parentActor = GetOwner();
	parentActor->OnActorBeginOverlap.AddDynamic(this, &UWeArtTouchableObject::BeginOverlap);
	parentActor->OnActorEndOverlap.AddDynamic(this, &UWeArtTouchableObject::EndOverlap);
}


// Called every frame
void UWeArtTouchableObject::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	// update volume texture
	weArtTexture.volume = VolumeTexture;
	
	if (grasperActor) {
		// Let the touchable object follow grasper actor if grabbed.
		GetOwner()->SetActorLocation(grasperActor->GetActorLocation());
		GetOwner()->SetActorRotation(grasperActor->GetActorRotation());
	}
	else {
		// Update all the touched effect every tick if not grabbed.
		for (std::pair<UWeArtHapticObject*, WeArtTouchEffect*> elem : touchedHapticEffects) {
			UWeArtHapticObject* obj = elem.first;
			WeArtTouchEffect* eff = elem.second;
			if (eff == nullptr)
				continue;

			WeArtImpactInfo impact = {
				true,
				obj->GetOwner()->GetActorLocation(),
				GetWorld()->GetTimeSeconds(),
				FMath::Clamp(CollisionMultiplier, WeArtConstants::minCollisionMultiplier, WeArtConstants::maxCollisionMultiplier)
			};

			bool changed = eff->Set(weArtTemperature, weArtStiffness, weArtTexture, impact);
			if (changed) {
				obj->UpdateEffects();
			}
		}
	}
}

// Event function bound to On Actor Begin Overlap
void UWeArtTouchableObject::BeginOverlap(AActor* MyOverlappedActor, AActor* OtherActor)
{
	if (graspingState == GraspingState::Grabbed)
		return;

	UWeArtHapticObject* hapticObject = TryGetHapticObjectFromActor(OtherActor);
	if (!hapticObject)
		return;

	touchedHapticObjects.insert(hapticObject);

	if (grasperActor)
		return;

	WeArtTouchEffect* eff = touchedHapticEffects[hapticObject];
	if (eff)
	{
		bool changed = eff->Set(weArtTemperature, weArtStiffness, weArtTexture, WeArtImpactInfo::none);
		if (changed)
			hapticObject->UpdateEffects();
	}
	else {
		eff = new WeArtTouchEffect(weArtTemperature, weArtStiffness, weArtTexture);
		hapticObject->AddEffect(eff);
		touchedHapticEffects[hapticObject] = eff;
	}

	
}

// Event function bound to On Actor End Overlap
void UWeArtTouchableObject::EndOverlap(AActor* MyOverlappedActor, AActor* OtherActor)
{
	if (graspingState == GraspingState::Grabbed)
		return;

	UWeArtHapticObject* hapticObject = TryGetHapticObjectFromActor(OtherActor);
	if (!hapticObject)
		return;

	touchedHapticObjects.erase(hapticObject);

	if (grasperActor)
		return;

	WeArtTouchEffect* eff = touchedHapticEffects[hapticObject];
	if (!eff)
		return;

	hapticObject->RemoveEffect(eff);
	touchedHapticEffects.erase(hapticObject);
	delete eff;
}

void UWeArtTouchableObject::UpdateTemperature(bool active, float value) {
	weArtTemperature.active = active;
	weArtTemperature.value = FMath::Clamp(value, 0.0f, 1.0f);
}

void UWeArtTouchableObject::UpdateStiffness(bool active, float value) {
	weArtStiffness.active = active;
	weArtStiffness.value = FMath::Clamp(value, 0.0f, 1.0f);;
}

void UWeArtTouchableObject::UpdateTexture(bool active, TextureType textureType) {
	weArtTexture.active = active;
	weArtTexture.textureType = textureType;
}

void UWeArtTouchableObject::UpdateTextureVolume(float value)
{
	VolumeTexture = FMath::Clamp(value, 0.0f, 100.0f);
}

void UWeArtTouchableObject::UpdateIsForcedVelocity(bool value)
{
	weArtTexture.isForcedTextureVelocity = value;
}

float UWeArtTouchableObject::GetTemperatureValue() {
	return weArtTemperature.value;
}

float UWeArtTouchableObject::GetStiffnessValue() {
	return weArtStiffness.value;
}

TextureType  UWeArtTouchableObject::GetTextureType() {
	return weArtTexture.textureType;
}

float UWeArtTouchableObject::GetTextureVolume() {
	return weArtTexture.volume;
}

bool UWeArtTouchableObject::GetIsForcedVelocity()
{
	return weArtTexture.isForcedTextureVelocity;
}


void UWeArtTouchableObject::Grab(AActor* grasper, UWeArtHapticObject* thumbHaptic, UWeArtHapticObject* indexHaptic, UWeArtHapticObject* middleHaptic)
{
	graspingState = GraspingState::Grabbed;

	grasperActor = grasper;

	UPrimitiveComponent* primitive = Cast<UPrimitiveComponent>(GetOwner()->GetRootComponent());
	isSimulatingPhysics = primitive->IsSimulatingPhysics();
	primitive->SetSimulatePhysics(false);

	thumbHapticObject = thumbHaptic;
	indexHapticObject = indexHaptic;
	middleHapticObject = middleHaptic;

	SetEffectGraspForce(WeArtConstants::defaultGraspForce, WeArtConstants::defaultGraspForce, WeArtConstants::defaultGraspForce);
}

void UWeArtTouchableObject::Release(void)
{
	grasperActor = nullptr;

	UPrimitiveComponent* primitive = Cast<UPrimitiveComponent>(GetOwner()->GetRootComponent());
	primitive->SetSimulatePhysics(isSimulatingPhysics);

	SetEffectGraspForce(0.0f, 0.0f, 0.0f);
	thumbHapticObject->RemoveEffect(thumbGraspEffect);
	indexHapticObject->RemoveEffect(indexGraspEffect);
	middleHapticObject->RemoveEffect(middleGraspEffect);

	thumbHapticObject = nullptr;
	indexHapticObject = nullptr;
	middleHapticObject = nullptr;

	if (touchedHapticObjects.find(thumbHapticObject) == touchedHapticObjects.end())
	{
		WeArtTouchEffect* eff = touchedHapticEffects[thumbHapticObject];
		if (eff)
		{
			thumbHapticObject->RemoveEffect(eff);
			touchedHapticEffects.erase(thumbHapticObject);
		}
	}

	if (touchedHapticObjects.find(indexHapticObject) == touchedHapticObjects.end())
	{
		WeArtTouchEffect* eff = touchedHapticEffects[indexHapticObject];
		if (eff)
		{
			indexHapticObject->RemoveEffect(eff);
			touchedHapticEffects.erase(indexHapticObject);
		}
	}

	if (touchedHapticObjects.find(middleHapticObject) == touchedHapticObjects.end())
	{
		WeArtTouchEffect* eff = touchedHapticEffects[middleHapticObject];
		if (eff)
		{
			middleHapticObject->RemoveEffect(eff);
			touchedHapticEffects.erase(middleHapticObject);
		}
	}

	graspingState = GraspingState::Released;
}


void UWeArtTouchableObject::SetEffectGraspForce(float thumbForce, float indexForce, float middleForce)
{
	FWeArtForce force;
	force.active = true;

	FWeArtTexture noTexture = FWeArtTexture();
	noTexture.active = false;

	force.value = thumbForce;
	if (thumbGraspEffect == nullptr)
		thumbGraspEffect = new WeArtTouchEffect(weArtTemperature, force, noTexture);

	if (!thumbHapticObject->ContainsEffect(thumbGraspEffect))
		thumbHapticObject->AddEffect(thumbGraspEffect);
	else
	{
		bool changed = thumbGraspEffect->Set(weArtTemperature, force, noTexture, WeArtImpactInfo::none);
		if (changed)
			thumbHapticObject->UpdateEffects();
	}


	force.value = indexForce;
	if (indexGraspEffect == nullptr)
		indexGraspEffect = new WeArtTouchEffect(weArtTemperature, force, noTexture);

	if (!indexHapticObject->ContainsEffect(indexGraspEffect))
		indexHapticObject->AddEffect(indexGraspEffect);
	else
	{
		bool changed = indexGraspEffect->Set(weArtTemperature, force, noTexture, WeArtImpactInfo::none);
		if (changed)
			indexHapticObject->UpdateEffects();
	}


	force.value = middleForce;
	if (middleGraspEffect == nullptr)
		middleGraspEffect = new WeArtTouchEffect(weArtTemperature, force, noTexture);

	if (!middleHapticObject->ContainsEffect(middleGraspEffect))
		middleHapticObject->AddEffect(middleGraspEffect);

	else
	{
		bool changed = middleGraspEffect->Set(weArtTemperature, force, noTexture, WeArtImpactInfo::none);
		if (changed)
			middleHapticObject->UpdateEffects();
	}

}


void UWeArtTouchableObject::ComputeDynamicGraspForce(float thumbClosureValue, float middleClosureValue, float indexClosureValue)
{
	const float deltaThumb = thumbClosureValue - WeArtConstants::thresholdThumbClosure;
	const float deltaMiddle = middleClosureValue - WeArtConstants::thresholdMiddleClosure;
	const float deltaIndex = indexClosureValue - WeArtConstants::thresholdIndexClosure;

	const float stiffnessObject = weArtStiffness.value;

	float dynamicForceThumb = (deltaThumb * stiffnessObject) * WeArtConstants::dynamicForceSensibility;
	float dynamicForceMiddle = (deltaMiddle * stiffnessObject) * WeArtConstants::dynamicForceSensibility;
	float dynamicForceIndex = (deltaIndex * stiffnessObject) * WeArtConstants::dynamicForceSensibility;

	dynamicForceThumb = FMath::Clamp(dynamicForceThumb, WeArtConstants::defaultGraspForce, 1.0f);
	dynamicForceIndex = FMath::Clamp(dynamicForceIndex, WeArtConstants::defaultGraspForce, 1.0f);
	dynamicForceMiddle = FMath::Clamp(dynamicForceMiddle, WeArtConstants::defaultGraspForce, 1.0f);

	SetEffectGraspForce(dynamicForceThumb, dynamicForceIndex, dynamicForceMiddle);
}
