#pragma once

#include "CoreMinimal.h"
#include "Components/ActorComponent.h"
#include "WeArtCommon.h"
#include "WeArtTouchEffect.h"
#include <set>
#include <map>

// Included cause of a bug in UnrealHeaderTool.
// Those are already included in WeArtEffect.h,
// but need to be included there as well.
#include "WeArtTemperature.h"
#include "WeArtForce.h"
#include "WeArtTexture.h"
#include "WeArtHapticObject.h"
#include "WeArtTouchableObject.generated.h"


// A class representing objects providing nontrivial effect to the player's
// @HapticObject. Interacts with the Haptic Objects by adding effect on begin
// overlap and removing then on overlap end.
// The effects themselves are then handled by the Haptic Object itself.
UCLASS( ClassGroup=(Custom), Blueprintable, meta=(BlueprintSpawnableComponent) )
class WEARTPLUGIN_API UWeArtTouchableObject : public UActorComponent
{
	GENERATED_BODY()

public:
	// Sets default values for this component's properties
	UWeArtTouchableObject();

protected:

	// State variables
	UPROPERTY(EditAnywhere, Meta = (DisplayName = "Temperature"))
		FWeArtTemperature weArtTemperature;
	UPROPERTY(EditAnywhere, Meta = (DisplayName = "Stiffness"))
		FWeArtForce weArtStiffness;
	UPROPERTY(EditAnywhere, Meta = (DisplayName = "Texture"))
		FWeArtTexture weArtTexture;

	UPROPERTY(EditAnywhere, Meta = (DisplayName = "Volume Texture", ClampMin = 0.0, ClampMax = 100.0))
		float VolumeTexture = WeArtConstants::defaultVolumeTexture;

	// Grasping-related variables.
	AActor* grasperActor = nullptr;
	UWeArtHapticObject* thumbHapticObject;
	UWeArtHapticObject* indexHapticObject;
	UWeArtHapticObject* middleHapticObject;

	// Called when the game starts
	virtual void BeginPlay() override;

public:
	bool isSimulatingPhysics;
	GraspingState graspingState = GraspingState::Released;

	UPROPERTY(EditAnywhere, Meta = (DisplayName = "Graspable"))
		bool IsGraspable;

	UPROPERTY(NoClear, Meta = (DisplayName = "Collision Multiplier", ClampMin = 0.0, ClampMax = 100.0))
		float CollisionMultiplier = WeArtConstants::defaultCollisionMultiplier;

	// Called every frame
	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;

	// Overlap handlers. Unity's equivalent of "StayOverlapped" is handled by the tick event.
	// Hit events are not currently handled at all, being more specific than their Unity counterparts
	// and not really suitable for our use.
	UFUNCTION()
		void BeginOverlap(AActor* MyOverlappedActor, AActor* OtherActor);
	UFUNCTION()
		void EndOverlap(AActor* MyOverlappedActor, AActor* OtherActor);

	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		void UpdateTemperature(bool active, float value);
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		void UpdateStiffness(bool active, float value);
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		void UpdateTexture(bool active, TextureType textureType);

	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		float GetTemperatureValue();
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		float GetStiffnessValue();
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		TextureType GetTextureType();
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		float GetTextureVolume();
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		void UpdateIsForcedVelocity(bool value);
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		bool GetIsForcedVelocity();
	UFUNCTION(BlueprintCallable, Category = "WEART TouchableObject")
		void UpdateTextureVolume(float value);

	void Grab(AActor* grasper, UWeArtHapticObject* thumbHaptic, UWeArtHapticObject* indexHaptic, UWeArtHapticObject* middleHaptic);
	void Release(void);
	void ComputeDynamicGraspForce(float thumbClosureValue, float middleClosureValue, float indexClosureValue);
	void SetEffectGraspForce(float thumbForce, float indexForce, float middleForce);



protected:
	std::set<UWeArtHapticObject*> touchedHapticObjects;
	std::map<UWeArtHapticObject*, WeArtTouchEffect*> touchedHapticEffects;

private:
	WeArtTouchEffect* thumbGraspEffect = nullptr;
	WeArtTouchEffect* indexGraspEffect = nullptr;
	WeArtTouchEffect* middleGraspEffect = nullptr;
};
